#include <avr/io.h>
#include <stdint.h>
#include <avr/interrupt.h>


#include "i2c.h"


/*
 * These codes are from the atmega8 datasheet
 */

#define TWSR_START_OK			0x08
#define TWSR_RESTART_OK			0x10
#define TWSR_SLA_W_TX_ACK		0x18
#define TWSR_SLA_W_TX_NACK		0x20
#define TWSR_DATA_TX_ACK		0x28
#define TWSR_DATA_TX_NACK		0x30
#define TWSR_ARBRITATION_LOST	0x38
#define TWSR_SLA_R_TX_ACK		0x40
#define TWSR_SLA_R_TX_NACK		0x48
#define TWSR_DATA_RX_ACK		0x50
#define TWSR_DATA_RX_NACK		0x58
#define TWSR_MASK				0xF8


void i2c_init(void)
{
	TWBR = 255;									/* Set some speed */
}


uint8_t i2c_start(void)
{
	TWCR = (1<<TWSTA) | (1<<TWINT) | (1<<TWEN);	/* Generate start condition */
	while(! (TWCR & (1<<TWINT)));				/* Wait until int flag set */
	if((TWSR & TWSR_MASK) != TWSR_START_OK) 	/* Check if start transmitted ok */
		return I2C_ERR;

	return I2C_OK;
}


uint8_t i2c_restart(void)
{
	TWCR = (1<<TWSTA) | (1<<TWINT) | (1<<TWEN);	/* Generate start condition */
	while(! (TWCR & (1<<TWINT)));				/* Wait until int flag set*/
	if((TWSR & TWSR_MASK) != TWSR_RESTART_OK) 	/* Check if restart transmitted ok */
		return I2C_ERR;

	return I2C_OK;
}


uint8_t i2c_stop(void)
{
	TWCR = (1<<TWSTO) | (1<<TWINT) | (1<<TWEN);	/* Generate stop condition */
	while((TWCR & (1<<TWSTO)));					/* Wait until int stop done */

	return I2C_OK;
}


uint8_t i2c_write(uint8_t data)
{
	TWDR = data;								/* Load data in TX register */
	TWCR = (1<<TWINT) | (1<<TWEN);				/* Initiate data send */
	while(! (TWCR & (1<<TWINT)));				/* Wait until int flag set */
	
	switch( TWSR & TWSR_MASK ) {				/* Check twi status code */
		case TWSR_DATA_TX_ACK:					/* These are various status codes */
		case TWSR_SLA_W_TX_ACK:					/*   which are OK, depending on */
		case TWSR_SLA_R_TX_ACK:					/*   the exact type of data sent */
			return I2C_OK;
			break;
	}
	return I2C_ERR;
}


uint8_t i2c_read(uint8_t *data)
{
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);	/* Initiate data receive, _with_ ack */
	while(! (TWCR & (1<<TWINT)));				/* Wait until int flag set */
	*data = TWDR;								/* Get received data */
	if((TWSR & TWSR_MASK) != TWSR_DATA_RX_ACK) 	/* Check if status code ok */
		return I2C_ERR;
		
	return I2C_OK;
}


uint8_t i2c_read_last(uint8_t *data)
{
	TWCR = (1<<TWINT) | (1<<TWEN);				/* Initiate data receive, _without_ ack */
	while(! (TWCR & (1<<TWINT)));				/* Wait until int flag set */
	*data = TWDR;								/* Get received data */
	if((TWSR & TWSR_MASK) != TWSR_DATA_RX_NACK) /* Check if status code ok */
		return I2C_ERR;
		
	return I2C_OK;
}

/* 
 * End 
 */
