#include <stdint.h> 
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include "uart.h"

FILE *f_uart0;
FILE *f_uart1;

#define RB_SIZE 64

struct ringbuffer_t {
	uint8_t buf[RB_SIZE];
	uint8_t head;
	uint8_t tail;
};

static int uart_put(char c, FILE *f);

static volatile struct ringbuffer_t rb_tx[2];
static volatile struct ringbuffer_t rb_rx[2];

static FILE my_uart0 = FDEV_SETUP_STREAM(uart_put, NULL, _FDEV_SETUP_WRITE);
static FILE my_uart1 = FDEV_SETUP_STREAM(uart_put, NULL, _FDEV_SETUP_WRITE);

/*
 * Initialize the UART. Should be called before any of the other uart functions is used
 */

static void uart_setup(enum uart_id id, uint16_t baudrate_divider, uint8_t open_stdout)
{
	volatile struct ringbuffer_t *rb;
	
	rb = &rb_tx[id];
	rb->head = rb->tail = 0;
	
	if(id == UART0) {
		UBRR0H = (baudrate_divider >> 8);				// Set baudrate
		UBRR0L = (baudrate_divider & 0xff);			
		UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);				// Set to N81
		UCSR0B = (1<<RXCIE0) | (1<<RXEN0) | (1<<TXEN0);	// Enable rx and tx interrupts
		f_uart0 = &my_uart0;
		if(open_stdout) stdout = f_uart0;
	}

	if(id == UART1) {
		UBRR1H = (baudrate_divider >> 8);				// Set baudrate
		UBRR1L = (baudrate_divider & 0xff);			
		UCSR1C = (1<<UCSZ11) | (1<<UCSZ10);				// Set to N81
		UCSR1B = (1<<RXCIE1) | (1<<RXEN1) | (1<<TXEN1);	// Enable rx and tx interrupts
		f_uart1 = &my_uart1;
		if(open_stdout) stdout = f_uart1;
	}
}


void uart_init(void)
{
	uart_setup(UART0, UART_BAUD(9600), 1);
	uart_setup(UART1, UART_BAUD(9600), 0);
}


/*
 * Write one character to the transmit ringbuffer
 */ 
 
void uart_tx(enum uart_id id, uint8_t c)
{
	volatile struct ringbuffer_t *rb = &rb_tx[id];
	uint8_t head;
	head = (rb->head + 1) % RB_SIZE;
	
	while(rb->tail == head);		/* Wait for free space in the buffer */
	rb->buf[rb->head] = c;			/* Put the char in the ringbuffer */
	rb->head = head;				/* Increase head ptr */
	
	if(id == UART0) {
		UCSR0B |= (1<<UDRIE0);		/* Enable TX-reg empty isr */
	} else {
		UCSR1B |= (1<<UDRIE1);		/* Enable TX-reg empty isr */
	}
}

uint8_t uart_rx(enum uart_id id)
{
	volatile struct ringbuffer_t *rb = &rb_rx[id];
	uint8_t c = rb->buf[rb->tail];
	rb->tail = (rb->tail + 1) % RB_SIZE;
	return c;
}

uint8_t uart_rx_avail(enum uart_id id)
{
	volatile struct ringbuffer_t *rb = &rb_rx[id];
	return(rb->head != rb->tail);
}

static int uart_put(char c, FILE *f)
{
	if(f == &my_uart0) uart_tx(UART0, c);
	if(f == &my_uart1) uart_tx(UART1, c);
	return 0;
}

ISR(USART0_RX_vect)
{
	volatile struct ringbuffer_t *rb = &rb_rx[0];
	rb->buf[rb->head] = UDR0;
	rb->head = (rb->head + 1) % RB_SIZE;
}

ISR(USART1_RX_vect)
{
	volatile struct ringbuffer_t *rb = &rb_rx[1];
	rb->buf[rb->head] = UDR1;
	rb->head = (rb->head + 1) % RB_SIZE;
}

/*
 * TX-register empty interrupt 
 */

ISR(USART0_UDRE_vect)
{
	volatile struct ringbuffer_t *rb = &rb_tx[0];
	
	UDR0 = rb->buf[rb->tail];			/* Put char in TX reg */
	rb->tail = (rb->tail + 1) % RB_SIZE;
	
	if(rb->tail == rb->head) {			/* If rb empty, disable irq */
		UCSR0B &= ~(1<<UDRIE0);
	}
}

ISR(USART1_UDRE_vect)
{
	volatile struct ringbuffer_t *rb = &rb_tx[1];
	
	UDR1 = rb->buf[rb->tail];			/* Put char in TX reg */
	rb->tail = (rb->tail + 1) % RB_SIZE;
	
	if(rb->tail == rb->head) {			/* If rb empty, disable irq */
		UCSR1B &= ~(1<<UDRIE1);
	}
}

/*
 * End
 */

